
#include <windows.h>
#include "lua.h"
#include "luafiber.h"

typedef struct lfib_data lfib_data;

enum lfib_state
{
	LFIB_SUSPENDED,
	LFIB_ACTIVE,
	LFIB_DEAD
};

struct lfib_data
{
	lua_State *thread;
	int state;
	int nargs;
	void *fiber;
	void *calling_fiber;
};

static char fiber_stash[] = "fiber_stash";

static void CALLBACK lfib_fiber_function(void *param)
{
	lfib_data *data = param;

	/* compute index of start function */
	int base = lua_gettop(data->thread) - data->nargs;

	/* call start function */
	int rc = lua_pcall(data->thread, data->nargs, LUA_MULTRET, 0);

	/* insert indicator before return values */
	lua_pushboolean(data->thread, rc == 0);
	lua_insert(data->thread, base);

	/* report total number of values available (including indicator) */
	data->nargs = lua_gettop(data->thread) - base + 1;

	/* we're done */
	data->state = LFIB_DEAD;

	/* finally switch back to caller */
	SwitchToFiber(data->calling_fiber);
}

static int lfib_gc(lua_State *L)
{
	lfib_data *data = lua_touserdata(L, 1);

	if (data->fiber)
	{
		DeleteFiber(data->fiber);
		data->fiber = 0;
	}

	return 0;
}

static int lfib_yield(lua_State *L)
{
	lfib_data *data = GetFiberData();

	/* insert indicator */
	lua_pushboolean(L, 1);
	lua_insert(L, 1);

	/* report number of values available */
	data->nargs = lua_gettop(L);

	/* report that we're suspended */
	data->state = LFIB_SUSPENDED;

	/* switch back to calling fiber */
	SwitchToFiber(data->calling_fiber);

	/* return the new set of values */
	return data->nargs;
}

static int lfib_resume(lua_State *L)
{
	lfib_data *data = lua_touserdata(L, 1);

	/* check fiber state */
	if (data->state != LFIB_SUSPENDED)
	{
		lua_pushboolean(L, 0);
		lua_pushstring(L, data->state == LFIB_ACTIVE ?
			"fiber already active" :
			"resuming a dead fiber");
		return 2;
	}

	/* remember which fiber to switch back to */
	data->calling_fiber = GetCurrentFiber();

	/* pass extra arguments to fiber */
	data->nargs = lua_gettop(L) - 1;
	lua_xmove(L, data->thread, data->nargs);

	/* activate... */
	data->state = LFIB_ACTIVE;
	SwitchToFiber(data->fiber);

	/* return the reported number of values */
	lua_xmove(data->thread, L, data->nargs);

	/* dead fiber? */
	if (data->state == LFIB_DEAD)
	{
		DeleteFiber(data->fiber);
		data->fiber = 0;

		/* remove from stash to enable gc of thread */
		lua_pushvalue(L, lua_upvalueindex(1));
		lua_pushvalue(L, 1);
		lua_pushnil(L);
		lua_rawset(L, -3);
		lua_pop(L, 1);
	}

	return data->nargs;
}

static int lfib_create(lua_State *L)
{
	lua_State *thread;
	lfib_data *data;

	/* trim stack to start function only */
	lua_settop(L, 1);

	/* create fiber */
	data = lua_newuserdata(L, sizeof(lfib_data));
	lua_pushvalue(L, lua_upvalueindex(2));
	lua_setmetatable(L, 2);

	/* create lua thread and stash it to prevent gc */
	lua_pushvalue(L, lua_upvalueindex(1));
	lua_pushvalue(L, 2);
	thread = lua_newthread(L);
	lua_rawset(L, -3);
	lua_pop(L, 1);

	/* prepare fiber */
	data->state = LFIB_SUSPENDED;
	data->thread = thread;
	data->fiber = CreateFiber(0, lfib_fiber_function, data);
	data->calling_fiber = 0;
	data->nargs = 0;

	/* move start function to new thread */
	lua_pushvalue(L, 1);
	lua_xmove(L, thread, 1);

	/* return fiber */
	return 1;
}

int luaopen_fiber(lua_State *L)
{
	/* enable fibers for calling thread */
	ConvertThreadToFiber(0);

	/* create fiber stash */
	lua_newtable(L);
	lua_newtable(L);
	lua_pushstring(L, "__mode");
	lua_pushstring(L, "k");
	lua_rawset(L, -3);
	lua_setmetatable(L, -2);

	/* place it in the registry */
	lua_pushlightuserdata(L, fiber_stash);
	lua_pushvalue(L, -2);
	lua_rawset(L, LUA_REGISTRYINDEX);

	/* create meta table */
	lua_newtable(L);
	lua_pushstring(L, "__gc");
	lua_pushcfunction(L, lfib_gc);
	lua_rawset(L, -3);

	/* create fiber global table */
	lua_newtable(L);

	/* add create function */
	lua_pushstring(L, "create");
	lua_pushvalue(L, -4);  /* fiber stash as first upvalue */
	lua_pushvalue(L, -4);  /* meta table as second upvalue */
	lua_pushcclosure(L, lfib_create, 2);
	lua_rawset(L, -3);

	lua_pushstring(L, "yield");
	lua_pushcfunction(L, lfib_yield);
	lua_rawset(L, -3);

	lua_pushstring(L, "resume");
	lua_pushvalue(L, -4);  /* fiber stash as first upvalue */
	lua_pushcclosure(L, lfib_resume, 1);
	lua_rawset(L, -3);

	/* install global fiber table */
	lua_pushstring(L, "fiber");
	lua_pushvalue(L, -2);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pop(L, 3);
	return 0;
}
